package net.windwards.options;

/**
 * <p>
 * Superclass for dispatchers that can be fed to CommandOptionParser. For
 * simple use cases, you can use RegistryCommandDispatcher. The common use
 * case for implementing your own command dispatcher would be if you
 * cannot get a list of commands in advance to populate a
 * RegistryCommandDispatcher (e.g. if they are HTTP endpoints) or if
 * creating instances of each command is expensive.
 * </p>
 *
 * <p>
 * During parsing, the selected command will get all options and arguments
 * that occurs after the command name on the command line. In order to
 * receive additional arguments, the instance need to implement
 * AcceptsArgument.
 * </p>
 *
 * @param <C>  This is a type common to all commands.
 *
 * @see net.windwards.options.CommandOptionParser
 * @see net.windwards.options.RegistryCommandDispatcher
 */
public abstract class CommandDispatcher<C> implements AcceptsArguments {
    protected C command;

	public boolean handleArgument(String cmdname) {
		this.command = this.handleCommand(cmdname);
		return false;
	}

    /**
     * Once parsing is done, will return the instance of type C that was
     * produced by handleCommand.
     *
     * @return selected command.
     */
    public C getSelectedCommand() {
		return this.command;
	}

    OptionParser<C> getOptionParser() {
         return new OptionParser<C>(this.command);
    }

    /**
     * Implement this method to lookup and/or create an instance of
     * type C that represents the command that matches <i>command</i>.
     * Consider throwing {@link net.windwards.options.err.NoSuchCommandException}
     * if <i>command</i> does not match any known command.
     *
     * @param command Name of the command as typed by the user.
     * @return An instance of type C.
     */
	public abstract C handleCommand(String command);
}