package net.windwards.options;

import net.windwards.options.err.InvalidOptionValue;
import net.windwards.options.err.MandatoryValueMissing;
import net.windwards.options.err.NoSuchOptionException;
import net.windwards.options.err.UnknownCommandException;

/**
 * The CommandOptionParser supports the sort of aggregated commandline 
 * tools that provides an interface to multiple sub-commands.
 * 
 * <p><code>$ tool -v frobnicate -f /some/unfortunate/file</code>
 * 
 * <p>The CommandOptionParser itself handles any global options 
 * (such as -v in this example). The parser then considers the first 
 * non-option argument to be a command and calls the handleCommand() 
 * method of the CommandDispatcher instance, which is expected to return 
 * some object that represents the command. This object is passed to a new
 * OptionParser which process any \@Option annotations on the command
 * object. In this case -f. If the command object implements AcceptsArgument
 * it can receive file arguments, such as the filename in the example above.
 * </p>
 * <p>Usage may look something like this:
 * <p><pre>
 * interface Command { public void activate(); }
 * 
 * class Frobnicator implements Command { 
 *   \@Option(shortName="f") public boolean force = false;
 *   public void activate() { ... }
 * }
 * 
 * class Dispatcher extends CommandDispatcher&lt;Command&gt; {
 *   \@Option public boolean verbose = false;
 *   public Command handleCommand(String cmdName) {
 *     Command command = null;
 *     if (cmdName.equals("frobnicate")) {
 *       command = new Frobnicator(this);
 *     } else {
 *       throw new RuntimeException("Bad command: " + cmdName);
 *     }
 *     return command;
 *   }
 * }
 * Dispatcher dispatcher = new Dispatcher();
 * CommandOptionParser&lt;Dispatcher&gt; parser = new CommandOptionParser&lt;Dispatcher&gt;(dispatcher);
 * parser.parse(args);
 * Command command = dispatcher.getSelectedCommand();
 * command.activate();
 * <pre>
 * 
 * @author quest
 * @see net.windwards.options.examples.CommandExample
 * @param <OC>
 */
public class CommandOptionParser<OC extends CommandDispatcher> 
extends OptionParser<OC> {

	public CommandOptionParser(OC options) {
		super(options);
	}

	public String[] parse(String[] args) 
	throws NoSuchOptionException, MandatoryValueMissing, InvalidOptionValue {
		String[] remainder = super.parse(args);
		if (this.options.getSelectedCommand() == null) {
			throw new UnknownCommandException("");
		}
		return this.options.getOptionParser().parse(remainder);
	}
}
