package net.windwards.options;

import net.windwards.options.err.InvalidOptionValue;
import net.windwards.options.err.UnsupportedTypeException;
import net.windwards.options.valueparsers.BooleanValueParser;
import net.windwards.options.valueparsers.ClassValueParser;
import net.windwards.options.valueparsers.EnumValueParser;
import net.windwards.options.valueparsers.FileValueParser;
import net.windwards.options.valueparsers.NullValueParser;
import net.windwards.options.valueparsers.NumberValueParser;
import net.windwards.options.valueparsers.StringValueParser;
import net.windwards.options.valueparsers.URLValueParser;
import net.windwards.options.valueparsers.ValueParser;

import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Information about a single options as detected during scanning.
 *
 * <p>
 * Add a class reference to OptionDescription.parsers in order to
 * instruct OptionParser how to interpret option values into new types.
 * For example, OptionParser could be taught to understand BigInteger
 * thus:
 * </p>
 *
 * <pre>
 * OptionDescription.parsers.put(BigInteger.class, new ValueParser() {
 *   public Object parse(OptionDescription desc, Class&lt;?&gt; type, String value) throws InvalidOptionValue {
 *     return new BigInteger(value);
 *   }
 * });
 * </pre>
 *
 * <p>
 * Note that this occurs during scanning, so the new ValueParser must be
 * inserted before the OptionParser is created.
 * </p>
 *
 * @see net.windwards.options.OptionParser
 */
public class OptionDescription {
    /**
     * Map of known ValueParsers.
     */
    public static final Map<Class<?>, ValueParser> parsers = new HashMap<Class<?>, ValueParser>();

    static {
        ValueParser booleanParser = new BooleanValueParser();
        ValueParser numberParser = new NumberValueParser();

        parsers.put(boolean.class, booleanParser);
        parsers.put(Boolean.class, booleanParser);
        parsers.put(short.class, numberParser);
        parsers.put(Short.class, numberParser);
        parsers.put(int.class, numberParser);
        parsers.put(Integer.class, numberParser);
        parsers.put(long.class, numberParser);
        parsers.put(Long.class, numberParser);
        parsers.put(float.class, numberParser);
        parsers.put(Float.class, numberParser);
        parsers.put(double.class, numberParser);
        parsers.put(Double.class, numberParser);
        parsers.put(String.class, new StringValueParser());
        parsers.put(Class.class, new ClassValueParser());
        parsers.put(URL.class, new URLValueParser());
        parsers.put(File.class, new FileValueParser());
    }

    /**
	 * Full name of the option
	 */
	public String longOptionName;
	/**
	 * Short (one-letter) version of option, or null if none.
	 */
	public String shortOptionName;
	/**
	 * default value of this option at time of scanning
	 */
	public Object defaultValue;
	/**
	 * documentation string, as picked up from @Option
	 */
	public String documentation;
    /**
     * If now null, the values allowed for this option; currently used for enums
     */
    public Set<String> alternatives;
    /**
     * The field object representing the referenced class member
     */
    public Field field;

    public ValueParser valueParser;

    public OptionDescription(Field field) {
        Option annotation = field.getAnnotation(Option.class);

        this.field = field;
        this.longOptionName = field.getName();
        this.shortOptionName = annotation.shortName();
        this.documentation = annotation.documentation();
        Class<?> type = this.getType();
        Class<? extends ValueParser> parserKlazz = annotation.valueParser();
        if (parserKlazz != NullValueParser.class) {
            try {
                this.valueParser = parserKlazz.newInstance();
            } catch (Exception e) {
                throw new UnsupportedTypeException("Failed to instantiate " +
                        "value parser " + parserKlazz.getName() +
                        " for option --" + this.longOptionName);
            }
        } else if (type.isEnum()) {
            this.valueParser = new EnumValueParser();
            Object[] constants = type.getEnumConstants();
            this.alternatives = new HashSet<String>();
            for (Object enumValue : constants) {
                this.alternatives.add(enumValue.toString());
            }
        } else {
            this.valueParser = parsers.get(type);
            if(this.valueParser == null)
                throw new UnsupportedTypeException("Internal error: Option --" +
                        this.longOptionName + " has unsupported type " + type.getName());
        }
    }

    public void extractDefaultValue(Object options) {
        Class<?> type = field.getType();
        if (Collection.class.isAssignableFrom(type)) {
            type = (Class<?>)((ParameterizedType)field.getGenericType())
                    .getActualTypeArguments()[0];
            if(type == Boolean.class) {
                this.defaultValue = false; // unintuitive; this is inverted by setOption
            }
        } else {
            try {
                this.defaultValue = field.get(options);
            } catch (IllegalAccessException e) { /* No, not really */ }
        }
    }

    public Object interpretValue(String value) throws InvalidOptionValue {
        Class<?> type = this.getType();
        return this.valueParser.parse(this, type, value);
    }

    public Class<?> getType() {
        Class<?> type = field.getType();

        if (Collection.class.isAssignableFrom(type)) {
            type = (Class<?>)((ParameterizedType)field.getGenericType())
                    .getActualTypeArguments()[0];
        }

        return type;
    }

    public boolean equals(Object o) {
		if (o instanceof OptionDescription) {
			return ((OptionDescription)o).longOptionName.equals(this.longOptionName);
		} else {
			return false;
		}
	}
}