package net.windwards.options;

import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Arrays;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.windwards.options.err.ArgumentsNotExpectedException;
import net.windwards.options.err.BadCommandLineException;
import net.windwards.options.err.DuplicateOptionException;
import net.windwards.options.err.InvalidOptionValue;
import net.windwards.options.err.MandatoryValueMissing;
import net.windwards.options.err.NoSuchOptionException;

/**
 * <p>
 * OptionParser handles parsing of commandline options and arguments,
 * roughly adhering to the GNU standard.
 * </p>
 * <p>
 * Instantiate an OptionParser, parameterized by any object that has 
 * fields annotated with {@literal @}Option. Each such field will be
 * recognized as the long version of an option. E.g.:
 * </p>
 *
 * <pre>
 *   class MyOptions {
 *     {@literal @}Option public boolean verbose = false;
 *   }
 *   MyOptions options = new MyOptions();
 *   OptionParser&lt;MyOptions&gt; parser = new OptionParser&lt;MyOptions&gt;(options);
 * </pre>
 *
 * <p>
 * When OptionParser is instantiated, it will <i>scan</i> the supplied
 * object for annotated fields.
 *
 * There are several types recognized by the scanner:
 * Boolean/boolean, Short/short, Integer/int, Long/long, Float/float,
 * Double/double, String, File, URL, Class and enumerations.
 * You can also use Collections of these types if the
 * option can be used more than once. E.g.:
 * </p>
 *
 * <pre>
 *   {@literal @}Option public Set&lt;boolean&gt; verbose = new HashSet&lt;boolean&gt;();
 * </pre>
 *
 * <p>
 * would allow you to do something like "--verbose --verbose" and perhaps
 * write a helper like this:
 * </p>
 * 
 * <pre>
 *   public void log(int lvl, String message) {
 *     if (options.verbose.size() &gt;= lvl) {
 *       System.out.println(message);
 *     }
 *   }
 * </pre>
 * After the OptionParser has scanned the options instance, you can change
 * it. For example, you can load a properties file that configures 
 * options. After this, simply supply the parser with the commandline 
 * argument list:
 * <pre>
 *   parser.parse(args);
 * </pre>
 * Any options contained in args will now be set on options.
 * 
 * If your command accepts arguments, e.g. subcommands, files, et.c., 
 * your options instance need to implement the AcceptsArguments 
 * interface. The handleArgument method will be invoked for each 
 * argument encountered. E.g.:
 * <pre>
 *   class Options {
 *     {@literal @}Option public boolean verbose = false;
 *     
 *     Map&lt;String&gt; files = HashMap&lt;String&gt();
 *     public boolean handeArgument(String arg) {
 *       this.files.add(arg);
 *       return true;
 *     }
 *   }
 * </pre>
 * Returning true instructs OptionParser to continue parsing the 
 * commandline.
 * 
 * @see net.windwards.options.examples.SimpleExample
 *   
 * @author quest
 *
 * @param <OC> Any class. Should be the type of the object considered 
 *             for options.
 */
public class OptionParser<OC> {

	/**
	 * Unless you hold your own reference to the options instance, you 
	 * can access it from here.
	 */
	public OC options = null;
	/**
	 * descriptions hold a mapping of option names to OptionDescription 
	 * objects that contain metadata about the options found when 
	 * scanning the options instance.
	 */
	public Map<String, OptionDescription> descriptions;
	/**
	 * Mapping of short version of options to their respective 
	 * OptionDescription objects. Populated during OptionParser construction.
	 */
	public Map<String, OptionDescription> shortNames;

	public OptionParser(OC options) {
		this.options = options;
		this.populateDescriptions();
	}

	private Map<String, OptionDescription> populateDescriptions() {
		this.descriptions = scanFields(this.options.getClass(), this.options);
		this.shortNames = new HashMap<String, OptionDescription>();
		for(String key : this.descriptions.keySet()) {
			OptionDescription desc = this.descriptions.get(key);
			if (!desc.shortOptionName.equals("")) {
				if (this.shortNames.get(desc.shortOptionName) != null) {
					throw new DuplicateOptionException("More than one field " +
							"defines shortName " + desc.shortOptionName + ".");
				}
				this.shortNames.put(desc.shortOptionName, desc);
			}
		}
		return this.descriptions;
	}

	/**
	 * Scan a class for Options, to produce OptionDescriptions from a 
	 * class with @Option annotations. Normally, this is done on 
	 * construction of a new OptionParser instance, but you may want to 
	 * use this static method if you don't have an instance.
	 * 
	 * <p>NB: without an options instance, this method will not know 
	 * what the default values are.
	 * 
	 * @param klaz The class to scan for Options.
	 * @param options Instance of klaz to read @Options from.
	 * @return Mapping of option longname to OptionDescription.
	 */
    public static Map<String, OptionDescription> scanFields(Class<?> klaz, Object options) {
        Map<String, OptionDescription> descriptions =
                new HashMap<String, OptionDescription>();
        Set<Field> fields = new HashSet<Field>();
        collectFields(klaz, fields);
        for (Field field : fields) {
            field.setAccessible(true);
            Option annotation = field.getAnnotation(Option.class);
            if (annotation == null)
                continue;

            OptionDescription desc = new OptionDescription(field);
            // TODO It's a bit a difficult to get the default values
            // without an instance
            if (options != null)
                desc.extractDefaultValue(options);

            descriptions.put(field.getName(), desc);
        }
        return descriptions;
    }

    private static void collectFields(Class<?> klaz, Set<Field> fields) {
        Class<?> parent = klaz.getSuperclass();
        if (parent != null && parent != Object.class)
            collectFields(parent, fields);
        fields.addAll(Arrays.asList(klaz.getDeclaredFields()));
    }

    private String stripQuotes(String string) {
		if (string == null) return null;
		if (string.startsWith("'") && string.endsWith("'")) 
			return string.substring(1, string.length() - 1);
		if (string.startsWith("\"") && string.endsWith("\""))
			return string.substring(1, string.length() - 1);
		return string;
	}

	/**
	 * Raw interface to set one option.
	 * 
	 * @param option  the name of the field on the options instance.
	 * @param value   the value in String form
	 * @throws NoSuchOptionException If setOption does not recognize the option.
	 * @throws MandatoryValueMissing If the option requires a value but 
	 *         setOption is called with null.
	 * @throws InvalidOptionValue  If value is not acceptable for option.
	 */
	public void setOption(String option, String value) 
	throws MandatoryValueMissing, NoSuchOptionException, InvalidOptionValue {
		OptionDescription desc = this.descriptions.get(option);
		if (desc == null) {
			throw new NoSuchOptionException("Don't know about --" + 
					option + ".");
		}
		this.setOption(desc, value);		
	}
	
	@SuppressWarnings("unchecked")
	protected void setOption(OptionDescription desc, String value) 
	throws MandatoryValueMissing, InvalidOptionValue {
		if (desc.defaultValue == null && value == null) {
			throw new MandatoryValueMissing("Option --" + desc.longOptionName + 
					" requires a value.");
		}

		Class<?> type = desc.field.getType();
        Object valueObj = desc.interpretValue(value);
        try {
            if (Collection.class.isAssignableFrom(type)) {
                ((Collection<Object>)desc.field.get(this.options)).add(valueObj);
            } else {
                desc.field.set(this.options, valueObj);
            }
        } catch (IllegalAccessException e) { /* */ }
	}

	private OptionDescription lookupShortOption(String shortOptName) 
	throws NoSuchOptionException {
		OptionDescription desc = this.shortNames.get(shortOptName);
		if (desc == null)
			throw new NoSuchOptionException("Don't know -" + shortOptName + ".");
		return desc;
	}

	protected String processShortOptions(String options, String value) 
	throws NoSuchOptionException, MandatoryValueMissing, InvalidOptionValue {
		assert options.charAt(0) == '-';
		String shortOpt;
		for(int i = 1; i < options.length() - 1; i++) {
			shortOpt = options.substring(i, i + 1);
			this.setOption(this.lookupShortOption(shortOpt), null);
		}
		shortOpt = options.substring(options.length() -1);
		OptionDescription desc = this.lookupShortOption(shortOpt);
		if (desc.defaultValue != null) {
			// short option booleans can't have values
			if (desc.defaultValue.getClass() == boolean.class || 
					desc.defaultValue.getClass() == Boolean.class) {
				this.setOption(desc, null);
				return value;
			} 
		}
		this.setOption(desc, value);
		return null;
	}
	
	protected Pattern longOptPattern = Pattern.compile("--([A-Za-z0-9]+)(=(.*))?");

	protected void processLongOption(String arg) 
	throws NoSuchOptionException, MandatoryValueMissing, InvalidOptionValue {
		Matcher match = this.longOptPattern.matcher(arg);
		if (match.matches()) {
			String optionName = match.group(1);
			String value = null;
			if (match.groupCount() > 1) 
				value = this.stripQuotes(match.group(3));
			this.setOption(optionName, value);
		} else {
			throw new BadCommandLineException("I cannot parse arg " + 
					arg + ".");
		}
	}

	/**
	 * Loads a Properties instance into the options instance given to 
	 * this OptionParser. You can use this to load default values 
	 * before calling parse().
	 * 
	 * <p>Any options named in the Properties file that don't match 
	 * fields on this.options are ignored.
	 * 
	 * @param properties a java.util.Properties instance.
	 * @throws InvalidOptionValue 
	 */
	public void loadProperties(Properties properties) throws InvalidOptionValue {
		Enumeration<Object> propenum = properties.keys();
		while(propenum.hasMoreElements()) {
			String key = (String)propenum.nextElement();
			String value = properties.getProperty(key);
			
			try {
				this.setOption(key, value);
			} catch (MandatoryValueMissing e) { /* No, not really */
			} catch (NoSuchOptionException e) { /* This is ok. */ }
		}
	}
	
	/**
	 * Parse a String[] of commandline arguments and populate the 
	 * previously supplied options instance.
	 * 
	 * <p>Parsing is done roughly in accordance with GNU getopt 
	 * tradition. Thus, all the following are acceptable:
	 * 
	 * $ convert -fvvv --from=here -t there --force=on /some/file
	 * 
	 * One notable exception from getopt tradition is that OptionParser 
	 * will always consider "-pfoo" to be equal to "-p -f -o -o" no 
	 * matter what values these options may accept (whereas getopt may 
	 * consider this to be "-p foo").
	 * 
	 * @param args Typically as supplied to your main() method.
	 * @return Any remaining arguments, either because of -- or because 
	 *         an handleArgument said so.
	 * @throws InvalidOptionValue The value cannot be fitted to its option.
	 * @throws MandatoryValueMissing An option was incorrectly supplied 
	 * 		   without a value. E.g.: --foo should have been --foo=bar.
	 * @throws NoSuchOptionException Supplied option is not recognized.
	 */
	public String[] parse(String[] args) 
	throws NoSuchOptionException, MandatoryValueMissing, InvalidOptionValue {
		String shortOpt = null;
		for (int i = 0; i < args.length; i++) {
			String arg = this.stripQuotes(args[i]);
			if (shortOpt != null) {
				arg = this.processShortOptions(shortOpt, arg);
				shortOpt = null;
				if (arg == null) continue;
			}

			if (arg.equals("--")) {
                for (int j = i + 1; j < args.length; j++) {
                    arg = this.stripQuotes(args[j]);
                    if (!((AcceptsArguments)this.options).handleArgument(arg)) {
                        int rlength = args.length - (j + 1);
                        String[] remainder = new String[rlength];
                        System.arraycopy(args, j + 1, remainder, 0, rlength);
                        return remainder;
                    }
                }
                break;
			} else if (arg.startsWith("--")) {
				this.processLongOption(arg);
			} else if (arg.startsWith("-")) {
				shortOpt = arg;
			} else if (this.options instanceof AcceptsArguments) {
				if (!((AcceptsArguments)this.options).handleArgument(arg)) {
					int rlength = args.length - (i + 1);
					String[] remainder = new String[rlength];
					System.arraycopy(args, i + 1, remainder, 0, rlength);
					return remainder;
				}
			} else {
				throw new ArgumentsNotExpectedException("Arguments are " + 
						" not allowed (only options); " + arg + 
						" violates this.");
			}
		}
		if (shortOpt != null) this.processShortOptions(shortOpt, null);
		return new String[] {};
	}

	private static StringBuilder formatDocumentation(String doc) {
		StringBuilder formatted = new StringBuilder();
		formatted.append("   ");
		int rowlength = 4;
		for(String token : doc.split("\\s")) {
			if (rowlength > (72 - 24)  && rowlength + token.length() > 72) {
				formatted.append("\n    ");
				rowlength = 4;
			} else {
				formatted.append(" ");
				rowlength += 1;
			}
			formatted.append(token);
			rowlength += token.length();
		}
		formatted.append("\n");
		return formatted;
	}
	
	/**
	 * Static version of the usage message. Note that it will not 
	 * contain default values. 
	 * 
	 * @param klaz The class to pass to scanFields.
	 * @return String with a formatted message of available options.
	 */
	public static String usage(Class<?> klaz) {
		return usage(scanFields(klaz, null));
	}

	/**
	 * Produce a human-readable message describing the available options.
	 * 
	 * @return A String containing a formatted message.
	 */
	public String usage() {
		return usage(this.descriptions);
	}
	
	protected static String usage(Map<String, OptionDescription> descriptions) {
		StringBuilder usageMessage = new StringBuilder();
		for (OptionDescription desc : descriptions.values()) {
			usageMessage.append("--");
			usageMessage.append(desc.longOptionName);
			if (desc.defaultValue == null) {
				usageMessage.append("=<value>");
			}
			if (desc.shortOptionName.length() > 0) {
				usageMessage.append(" (-");
				usageMessage.append(desc.shortOptionName);
				usageMessage.append(")");
			}
			usageMessage.append("\n");
			if (desc.documentation.length() > 0) {
				usageMessage.append(formatDocumentation(desc.documentation));
			}
            if (desc.alternatives != null) {
                usageMessage.append("    Allowed values: ");
                boolean first = true;
                for (String alt : desc.alternatives) {
                    if (!first) usageMessage.append(", ");
                    first = false;
                    usageMessage.append(alt);
                }
                usageMessage.append(".\n");
            }
			if (desc.defaultValue != null) {
				usageMessage.append("    Default: ");
				usageMessage.append(desc.defaultValue);
				usageMessage.append("\n");
			}
		}
		return usageMessage.toString();
	}
}
